# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Standard flavors, flavor aliases, etc.... '''


from . import __
from . import compositors as _compositors
from . import core as _core
from . import introducers as _intros


def produce_flavors(
    introducercfg: _core.IntroducerConfiguration = (
        _core.INTRODUCER_CONFIGURATION_DEFAULT ),
    compositorcfg: _core.CompositorConfiguration = (
        _core.COMPOSITOR_CONFIGURATION_DEFAULT ),
) -> __.FlavorsRegistry:
    ''' Produces registry of all standard flavors.

        Customization of introducers and compositors is possible.
    '''
    flavors: __.FlavorsRegistryLiberal = { }
    introducer = _intros.Introducer( configuration = introducercfg )
    compositor = _compositors.Compositor(
        configuration = compositorcfg, introducer = introducer )
    for name, spec in __.flavor_specifications_standard.items( ):
        ccfg = compositorcfg
        if spec.stack:
            lcfg = ccfg.linearizercfg
            ecfg = __.dcls.replace(
                lcfg.exceptionscfg,
                enable_discovery = True,
                enable_stacktraces = True )
            lcfg = __.dcls.replace( lcfg, exceptionscfg = ecfg )
            ccfg = __.dcls.replace( ccfg, linearizercfg = lcfg )
        compositor = _compositors.Compositor(
            configuration = ccfg, introducer = introducer )
        flavors[ name ] = __.FlavorConfiguration(
            compositor_factory = _produce_compositor_factory( compositor ) )
    for alias, name in __.flavor_aliases_standard.items( ):
        flavors[ alias ] = flavors[ name ]
    for level in range( 10 ):
        indent_i = '  ' * level
        indent_s = '  ' * ( level + 1 )
        ccfg = __.dcls.replace(
            compositorcfg,
            line_prefix_initial = indent_i,
            line_prefix_subsequent = indent_s )
        compositor = _compositors.Compositor(
            configuration = ccfg, introducer = introducer )
        flavors[ level ] = __.FlavorConfiguration(
            compositor_factory = _produce_compositor_factory( compositor ) )
    return __.immut.Dictionary( flavors )


def _produce_compositor_factory(
    compositor: _compositors.Compositor
) -> __.CompositorFactory:
    return lambda address, flavor: compositor



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Interfaces for compositors, introducers, et cetera. '''


from . import __
from . import flavors as _flavors
from . import printers as _printers
from . import records as _records


class Compositor( __.immut.DataclassProtocol, __.typx.Protocol ):
    ''' Abstract base class for compositors. '''

    @__.abc.abstractmethod
    def __call__(
        self,
        control: _printers.TextualizationControl,
        record: _records.Record,
    ) -> str:
        ''' Renders record as text. '''
        raise NotImplementedError


CompositorFactory: __.typx.TypeAlias = (
    __.typx.Callable[ [ str, _flavors.Flavor ], Compositor ] )


class Introducer( __.immut.DataclassProtocol, __.typx.Protocol ):
    ''' Abstract base class for introducers. '''

    @__.abc.abstractmethod
    def __call__(
        self,
        control: _printers.TextualizationControl,
        record: _records.Record,
        columns_max: __.Absential[ int ] = __.absent,
    ) -> str:
        ''' Renders record as text label. '''
        raise NotImplementedError


IntroducerUnion: __.typx.TypeAlias = str | Introducer


class Linearizer( __.immut.DataclassProtocol, __.typx.Protocol ):
    ''' Abstract base class for linearizers. '''

    @__.abc.abstractmethod
    def __call__(
        self,
        control: _printers.TextualizationControl,
        entity: object,
        columns_max: __.Absential[ int ] = __.absent,
    ) -> tuple[ str, ... ]:
        ''' Renders object as lines of text. '''
        raise NotImplementedError


def produce_compositor_factory_default(
    introducer: __.Absential[ IntroducerUnion ] = __.absent,
    line_prefix_initial: str = '',
    line_prefix_subsequent: str = '  ',
    trace_exceptions: bool = False,
) -> CompositorFactory:
    ''' Produces default compositor factory. '''
    def produce_compositor(
        address: str, flavor: _flavors.Flavor
    ) -> Compositor:
        from .standard import (
            Compositor,
            CompositorConfiguration,
            ExceptionsConfiguration,
            LinearizerConfiguration,
        )
        ecfg = ExceptionsConfiguration(
            enable_discovery = trace_exceptions,
            enable_stacktraces = trace_exceptions )
        lcfg = LinearizerConfiguration( exceptionscfg = ecfg )
        ccfg = CompositorConfiguration(
            line_prefix_initial = line_prefix_initial,
            line_prefix_subsequent = line_prefix_subsequent,
            linearizercfg = lcfg )
        if __.is_absent( introducer ):
            return Compositor( configuration = ccfg )
        return Compositor( configuration = ccfg, introducer = introducer )

    return produce_compositor



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Message reporters. '''


from . import __
from . import flavors as _flavors
from . import printers as _printers
from . import records as _records
from . import textualizers as _texts


class Reporter( __.immut.DataclassObject ):
    ''' Formats and prints messages to targets. '''

    active: bool  # TODO? Also accept predicate function to decide if active.
    address: str
    flavor: _flavors.Flavor
    compositor: _texts.Compositor
    printers: _printers.Printers

    def __call__(
        self,
        summary: _records.MessageSummary, /,
        *details: _records.MessageDetail,
    ) -> None:
        # TODO? Return record.
        ''' Prepares record and prints it. '''
        if not self.active: return
        content = _records.MessageContent(
            summary = summary, details = details )
        record = _records.Record(
            address = self.address, content = content, flavor = self.flavor )
        for printer in self.printers:
            tcontrol = printer.provide_textualization_control( )
            if tcontrol is None: printer( record )
            else: printer( self.compositor( tcontrol, record ) )

    # TODO: inscribe (same as __call__)
    # TODO: inscribe_async
    # TODO? inspect
    # TODO? Ability to print stack traces either from current frame or from
    #       supplied traceback. Maybe various modes, such as compact or
    #       detailed (showing names and values of locals).



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Dispatchers for messages and inspections to scribes. '''


from . import __
from . import configuration as _cfg
from . import exceptions as _exceptions
from . import flavors as _flavors
from . import printers as _printers
from . import reporters as _reporters
from . import textualizers as _texts


_installer_mutex: __.threads.Lock = __.threads.Lock( )
_registrar_mutex: __.threads.Lock = __.threads.Lock( )
_self_addresscfg: _cfg.AddressConfiguration = _cfg.AddressConfiguration(
    flavors = __.immut.Dictionary(
        note = _cfg.FlavorConfiguration(
            compositor_factory = (
                _texts.produce_compositor_factory_default(
                    introducer = 'NOTE| ' ) ) ),
        error = _cfg.FlavorConfiguration(
            compositor_factory = (
                _texts.produce_compositor_factory_default(
                    introducer = 'ERROR| ' ) ) ) ) )
_validate_arguments = (
    __.validate_arguments(
        globalvars = globals( ),
        errorclass = _exceptions.ArgumentClassInvalidity ) )


class AddressesConfigurationsRegistry(
    __.accret.Dictionary[ str, _cfg.AddressConfiguration ]
):
    # TODO: Use 'accret.ValidatorDictionary'.
    ''' Accretive dictionary specifically for address registrations. '''

    def __init__(
        self,
        *iterables: __.DictionaryPositionalArgument[
            str, _cfg.AddressConfiguration ],
        **entries: __.DictionaryNominativeArgument[
            _cfg.AddressConfiguration ],
    ):
        super( ).__init__( { __.package_name: _self_addresscfg } )
        self.update( *iterables, **entries )


class Omniflavor( __.enum.Enum ):
    ''' Singleton to match any flavor. '''

    Instance = __.enum.auto( )


ActiveFlavors: __.typx.TypeAlias = Omniflavor | frozenset[ _flavors.Flavor ]
ActiveFlavorsLiberal: __.typx.TypeAlias = __.typx.Union[
    Omniflavor,
    __.cabc.Sequence[ _flavors.Flavor ],
    __.cabc.Set[ _flavors.Flavor ],
]
ActiveFlavorsRegistry: __.typx.TypeAlias = (
    __.immut.Dictionary[ __.typx.Optional[ str ], ActiveFlavors ] )
ActiveFlavorsRegistryLiberal: __.typx.TypeAlias = (
    __.cabc.Mapping[ __.typx.Optional[ str ], ActiveFlavorsLiberal ] )
AddressesConfigurationsRegistryLiberal: __.typx.TypeAlias = (
    __.cabc.Mapping[ str, _cfg.AddressConfiguration ] )
ReportersRegistry: __.typx.TypeAlias = (
    __.accret.Dictionary[
        tuple[ str, _flavors.Flavor ], _reporters.Reporter ] )
TraceLevelsRegistry: __.typx.TypeAlias = (
    __.immut.Dictionary[ __.typx.Optional[ str ], int ] )
TraceLevelsRegistryLiberal: __.typx.TypeAlias = (
    __.cabc.Mapping[ __.typx.Optional[ str ], int ] )


def _provide_active_flavors_default( ) -> ActiveFlavorsRegistry:
    ''' Provides default set of globally active flavors. '''
    flavors = set( _flavors.flavor_specifications_standard.keys( ) )
    flavors.update( _flavors.flavor_aliases_standard.keys( ) )
    return __.immut.Dictionary( { None: frozenset( flavors ) } )


builtins_alias_default: __.typx.Annotated[
    str,
    __.typx.Doc(
        ''' Default alias for global dispatcher in builtins module. ''' ),
] = 'ictr'
addresscfgs: __.typx.Annotated[
    AddressesConfigurationsRegistry,
    __.typx.Doc( ''' Global registry of address configurations. ''' ),
] = AddressesConfigurationsRegistry( )
omniflavor: __.typx.Annotated[
    Omniflavor, __.typx.Doc( ''' Matches any flavor. ''' )
] = Omniflavor.Instance


class Dispatcher( __.immut.DataclassObject ):
    ''' Provides reporter instances. '''

    active_flavors: __.typx.Annotated[
        ActiveFlavorsRegistry,
        __.typx.Doc(
            ''' Mapping of addresses to active flavor sets.

                Key ``None`` applies globally. Address-specific entries
                override globals for that address.
            ''' ),
    ] = __.dcls.field( default_factory = _provide_active_flavors_default )
    addresscfgs: __.typx.Annotated[
        AddressesConfigurationsRegistry,
        __.typx.Doc(
            ''' Registry of per-address configurations.

                Addresses inherit configuration from their parent packages.
                Top-level packages inherit from general instance
                configruration.
            ''' ),
    ] = __.dcls.field( default_factory = lambda: addresscfgs )
    generalcfg: __.typx.Annotated[
        _cfg.DispatcherConfiguration,
        __.typx.Doc(
            ''' General configuration.

                Top of configuration inheritance hierarchy.
                Default is suitable for application use.
            ''' ),
    ] = __.dcls.field( default_factory = _cfg.DispatcherConfiguration )
    printer_factories: __.typx.Annotated[
        _printers.PrinterFactoriesUnion,
        __.typx.Doc(
            ''' Factories which produce callables to output text somewhere.

                A factory takes two arguments, address and flavor, and
                returns a callable which takes one argument, either a record
                or the string produced by a textualizer.

                May also be writable text stream instead of a factory.
            ''' ),
    ] = ( _printers.produce_printer_factory_default( __.sys.stderr ), )
    reporters: __.typx.Annotated[
        ReportersRegistry,
        __.typx.Doc(
            ''' Cache of reporter instances by address and flavor. ''' ),
    ] = __.dcls.field( default_factory = ReportersRegistry )
    # TODO? Move reporters mutex into reporters registry.
    reporters_mutex: __.typx.Annotated[
        __.threads.Lock,
        __.typx.Doc( ''' Access lock for cache of reporter instances. ''' ),
    ] = __.dcls.field( default_factory = __.threads.Lock )
    trace_levels: __.typx.Annotated[
        TraceLevelsRegistry,
        __.typx.Doc(
            ''' Mapping of addresses to maximum trace depths.

                Key ``None`` applies globally. Address-specific entries
                override globals for that address.
            ''' ),
    ] = __.dcls.field(
        default_factory = lambda: __.immut.Dictionary( { None: -1 } ) )

    @_validate_arguments
    def __call__(
        self,
        flavor: _flavors.Flavor, *,
        address: __.Absential[ str ] = __.absent,
    ) -> _reporters.Reporter:
        ''' Produces and caches message reporter. '''
        address = (
            _discover_invoker_module_name( ) if __.is_absent( address )
            else address )
        cache_index = ( address, flavor )
        if cache_index in self.reporters:
            with self.reporters_mutex:
                return self.reporters[ cache_index ]
        configuration = _produce_ic_configuration( self, address, flavor )
        if isinstance( flavor, int ):
            trace_level = (
                _calculate_effective_trace_level( self.trace_levels, address) )
            active = flavor <= trace_level
        elif isinstance( flavor, str ): # pragma: no branch
            active_flavors = (
                _calculate_effective_flavors( self.active_flavors, address ) )
            active = (
                isinstance( active_flavors, Omniflavor )
                or flavor in active_flavors )
        compositor = configuration[ 'compositor_factory' ]( address, flavor )
        printers = _resolve_printers( self.printer_factories, address, flavor )
        reporter = _reporters.Reporter(
            active = active, address = address, flavor = flavor,
            compositor = compositor, printers = printers )
        with self.reporters_mutex:
            self.reporters[ cache_index ] = reporter
        return reporter

    @_validate_arguments
    def install( self, alias: str = builtins_alias_default ) -> __.typx.Self:
        ''' Installs dispatcher into builtins with provided alias.

            Replaces an existing dispatcher. Preserves global address
            configurations.

            Library developers should call :py:func:`register_address` instead.
        '''
        import builtins
        with _installer_mutex:
            dispatcher_o = getattr( builtins, alias, None )
            if isinstance( dispatcher_o, Dispatcher ):
                self( 'note', address = __name__ )(
                    "Installed dispatcher is being replaced." )
                setattr( builtins, alias, self )
            else:
                __.install_builtin_safely(
                    alias, self, _exceptions.AttributeNondisplacement )
        return self

    @_validate_arguments
    def register_address(
        self,
        name: __.Absential[ str ] = __.absent,
        configuration: __.Absential[ _cfg.AddressConfiguration ] = __.absent,
    ) -> __.typx.Self:
        ''' Registers configuration for address.

            If no address is given, then the invoking module name is
            inferred.

            If no configuration is provided, then a default is generated.
        '''
        if __.is_absent( name ):
            name = _discover_invoker_module_name( )
        if __.is_absent( configuration ):
            configuration = _cfg.AddressConfiguration( )
        with _registrar_mutex:
            self.addresscfgs[ name ] = configuration
        return self


ActiveFlavorsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ ActiveFlavorsLiberal | ActiveFlavorsRegistryLiberal ],
    __.typx.Doc(
        ''' Flavors to activate.

            Can be collection, which applies globally across all registered
            addresses. Or, can be mapping of addresses to sets.

            Address-specific entries merge with global entries.
        ''' ),
]
AddressArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ str ],
    __.typx.Doc(
        ''' Address to register.

            If absent, infers the invoking module name as the address.
        ''' ),
]
AddresscfgsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ AddressesConfigurationsRegistryLiberal ],
    __.typx.Doc(
        ''' Address configurations for the dispatcher.

            If absent, defaults to global addresses registry.
        ''' ),
]
CompositorFactoryArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _texts.CompositorFactory ],
    __.typx.Doc(
        ''' Factory which produces compositor callable.

            Takes address and flavor as arguments.
            Returns compositor to convert record content to a string.
        ''' ),
]
EvnActiveFlavorsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ __.typx.Optional[ str ] ],
    __.typx.Doc(
        ''' Name of environment variable for active flavors or ``None``.

            If absent, then a default environment variable name is used.

            If ``None``, then active flavors are not parsed from the process
            environment.

            If active flavors are supplied directly to a function,
            which also accepts this argument, then active flavors are not
            parsed from the process environment.
        ''' ),
]
EvnTraceLevelsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ __.typx.Optional[ str ] ],
    __.typx.Doc(
        ''' Name of environment variable for trace levels or ``None``.

            If absent, then a default environment variable name is used.

            If ``None``, then trace levels are not parsed from the process
            environment.

            If trace levels are supplied directly to a function,
            which also accepts this argument, then trace levels are not
            parsed from the process environment.
        ''' ),
]
FlavorsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _cfg.FlavorsRegistryLiberal ],
    __.typx.Doc( ''' Registry of flavor identifiers to configurations. ''' ),
]
GeneralcfgArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _cfg.DispatcherConfiguration ],
    __.typx.Doc(
        ''' General configuration for the dispatcher.

            Top of configuration inheritance hierarchy. If absent,
            defaults to a suitable configuration for application use.
        ''' ),
]
InstallAliasArgument: __.typx.TypeAlias = __.typx.Annotated[
    str,
    __.typx.Doc(
        ''' Alias under which the dispatcher is installed in builtins. ''' ),
]
IntroducerArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _texts.IntroducerUnion ],
    __.typx.Doc(
        ''' String or factory which produces message introduction.

            Factory takes textualization control, address, and flavor as
            arguments. Returns introduction string.
        ''' ),
]
PrinterFactoriesArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ _printers.PrinterFactoriesUnion ],
    __.typx.Doc(
        ''' Factories which produce callables to output text somewhere.

            A factory take two arguments, address and flavor, and
            returns a callable which takes one argument, either a record or
            the string produced by a textualizer.

            May also be writable text stream instead of a factory.

            If absent, uses a default.
        ''' ),
]
TraceLevelsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.Absential[ int | TraceLevelsRegistryLiberal ],
    __.typx.Doc(
        ''' Maximum trace depths.

            Can be an integer, which applies globally across all registered
            addresses. Or, can be a mapping of addresses to integers.

            Address-specific entries override global entries.
        ''' ),
]


def active_flavors_from_environment(
    evname: __.Absential[ str ] = __.absent
) -> ActiveFlavorsRegistry:
    ''' Extracts active flavors from named environment variable. '''
    active_flavors: ActiveFlavorsRegistryLiberal = { }
    name = 'ICTR_ACTIVE_FLAVORS' if __.is_absent( evname ) else evname
    value = __.os.getenv( name )
    if value is None:
        return _provide_active_flavors_default( )
    for part in value.split( '+' ):
        if not part: continue
        if ':' in part:
            address, flavors = part.split( ':', 1 )
        else: address, flavors = None, part
        match flavors:
            case '*': active_flavors[ address ] = omniflavor
            case _: active_flavors[ address ] = flavors.split( ',' )
    return __.immut.Dictionary( {
        address:
            flavors if isinstance( flavors, Omniflavor )
            else frozenset( flavors )
        for address, flavors in active_flavors.items( ) } )


def trace_levels_from_environment(
    evname: __.Absential[ str ] = __.absent
) -> TraceLevelsRegistry:
    ''' Extracts trace levels from named environment variable. '''
    trace_levels: TraceLevelsRegistryLiberal = { None: -1 }
    name = 'ICTR_TRACE_LEVELS' if __.is_absent( evname ) else evname
    value = __.os.getenv( name, '' )
    for part in value.split( '+' ):
        if not part: continue
        if ':' in part: address, level = part.split( ':', 1 )
        else: address, level = None, part
        if not level.isdigit( ):
            __.warnings.warn(
                f"Non-integer trace level {level!r} "
                f"in environment variable {name!r}." )
            continue
        trace_levels[ address ] = int( level )
    return __.immut.Dictionary( trace_levels )


@_validate_arguments
def install( # noqa: PLR0913
    alias: InstallAliasArgument = builtins_alias_default,
    active_flavors: ActiveFlavorsArgument = __.absent,
    generalcfg: GeneralcfgArgument = __.absent,
    printer_factories: PrinterFactoriesArgument = __.absent,
    trace_levels: TraceLevelsArgument = __.absent,
    evname_active_flavors: EvnActiveFlavorsArgument = __.absent,
    evname_trace_levels: EvnTraceLevelsArgument = __.absent,
) -> Dispatcher:
    ''' Produces dispatcher and installs it into builtins with alias.

        Replaces an existing dispatcher, preserving global address
        configurations.

        Library developers should call :py:func:`register_address` instead.
    '''
    dispatcher = produce_dispatcher(
        active_flavors = active_flavors,
        generalcfg = generalcfg,
        printer_factories = printer_factories,
        trace_levels = trace_levels,
        evname_active_flavors = evname_active_flavors,
        evname_trace_levels = evname_trace_levels )
    return dispatcher.install( alias = alias )


@_validate_arguments
def produce_dispatcher( # noqa: PLR0913
    active_flavors: ActiveFlavorsArgument = __.absent,
    generalcfg: GeneralcfgArgument = __.absent,
    addresscfgs: AddresscfgsArgument = __.absent,
    printer_factories: PrinterFactoriesArgument = __.absent,
    trace_levels: TraceLevelsArgument = __.absent,
    evname_active_flavors: EvnActiveFlavorsArgument = __.absent,
    evname_trace_levels: EvnTraceLevelsArgument = __.absent,
) -> Dispatcher:
    ''' Produces dispatcher with some shorthand argument values. '''
    # TODO: Deeper validation of active flavors and trace levels.
    # TODO: Deeper validation of printer factory.
    initargs: dict[ str, __.typx.Any ] = { }
    if not __.is_absent( generalcfg ):
        initargs[ 'generalcfg' ] = generalcfg
    if not __.is_absent( addresscfgs ):
        initargs[ 'addresscfgs' ] = AddressesConfigurationsRegistry(
            {   address: configuration for address, configuration
                in addresscfgs.items( ) } )
    if not __.is_absent( printer_factories ):
        initargs[ 'printer_factories' ] = printer_factories
    _add_dispatcher_initarg_active_flavors(
        initargs, active_flavors, evname_active_flavors )
    _add_dispatcher_initarg_trace_levels(
        initargs, trace_levels, evname_trace_levels )
    return Dispatcher( **initargs )


@_validate_arguments
def register_address(
    name: AddressArgument = __.absent,
    flavors: FlavorsArgument = __.absent,
    compositor_factory: CompositorFactoryArgument = __.absent,
    introducer: IntroducerArgument = __.absent,
) -> _cfg.AddressConfiguration:
    ''' Registers address configuration on the builtin dispatcher.

        If no dispatcher exists in builtins, installs one which produces null
        printers.

        Intended for library developers to configure debugging flavors
        without overriding anything set by the application or other libraries.
        Application developers should call :py:func:`install` instead.
    '''
    import builtins
    dispatcher = getattr( builtins, builtins_alias_default, None )
    if not isinstance( dispatcher, Dispatcher ):
        dispatcher = Dispatcher( )
        __.install_builtin_safely(
            builtins_alias_default,
            dispatcher,
            _exceptions.AttributeNondisplacement )
    nomargs: dict[ str, __.typx.Any ] = { }
    if not __.is_absent( flavors ):
        nomargs[ 'flavors' ] = __.immut.Dictionary( flavors )
    if not __.is_absent( compositor_factory ):
        nomargs[ 'compositor_factory' ] = compositor_factory
    if not __.is_absent( introducer ):
        nomargs[ 'introducer' ] = introducer
    configuration = _cfg.AddressConfiguration( **nomargs )
    dispatcher.register_address( name = name, configuration = configuration )
    return configuration


def _add_dispatcher_initarg_active_flavors(
    initargs: dict[ str, __.typx.Any ],
    active_flavors: ActiveFlavorsArgument = __.absent,
    evname_active_flavors: EvnActiveFlavorsArgument = __.absent,
) -> None:
    name = 'active_flavors'
    if not __.is_absent( active_flavors ):
        if isinstance( active_flavors, Omniflavor ):
            initargs[ name ] = __.immut.Dictionary(
                { None: active_flavors } )
        elif isinstance( active_flavors, ( __.cabc.Sequence,  __.cabc.Set ) ):
            initargs[ name ] = __.immut.Dictionary(
                { None: frozenset( active_flavors ) } )
        else:
            initargs[ name ] = __.immut.Dictionary( {
                address:
                    flavors if isinstance( flavors, Omniflavor )
                    else frozenset( flavors )
                for address, flavors in active_flavors.items( ) } )
    elif evname_active_flavors is not None:
        initargs[ name ] = (
            active_flavors_from_environment( evname = evname_active_flavors ) )


def _add_dispatcher_initarg_trace_levels(
    initargs: dict[ str, __.typx.Any ],
    trace_levels: TraceLevelsArgument = __.absent,
    evname_trace_levels: EvnTraceLevelsArgument = __.absent,
) -> None:
    name = 'trace_levels'
    if not __.is_absent( trace_levels ):
        if isinstance( trace_levels, int ):
            initargs[ name ] = __.immut.Dictionary( { None: trace_levels } )
        else:
            trace_levels_: TraceLevelsRegistryLiberal = { None: -1 }
            trace_levels_.update( trace_levels )
            initargs[ name ] = __.immut.Dictionary( trace_levels_ )
    elif evname_trace_levels is not None:
        initargs[ name ] = (
            trace_levels_from_environment( evname = evname_trace_levels ) )


def _calculate_effective_flavors(
    flavors: ActiveFlavorsRegistry, address: str
) -> ActiveFlavors:
    result_ = flavors.get( None ) or frozenset( )
    if isinstance( result_, Omniflavor ): return result_
    result = result_
    for address_ in _iterate_address_ancestry( address ):
        if address_ in flavors:
            result_ = flavors.get( address_ ) or frozenset( )
            if isinstance( result_, Omniflavor ): return result_
            result |= result_
    return result


def _calculate_effective_trace_level(
    levels: TraceLevelsRegistry, address: str
) -> int:
    result = levels.get( None, -1 )
    for address_ in _iterate_address_ancestry( address ):
        if address_ in levels:
            result = levels[ address_ ]
    return result


def _dict_from_dataclass( objct: object ) -> dict[ str, __.typx.Any ]:
    # objct = __.typx.cast( _typeshed.DataclassInstance, objct )
    return {
        field.name: getattr( objct, field.name )
        for field in __.dcls.fields( objct ) # pyright: ignore[reportArgumentType]
        if not field.name.startswith( '_' ) }


def _discover_invoker_module_name( ) -> str:
    frame = __.inspect.currentframe( )
    while frame: # pragma: no branch
        module = __.inspect.getmodule( frame )
        if module is None:
            if '<stdin>' == frame.f_code.co_filename: # pragma: no cover
                name = '__main__'
                break
            raise _exceptions.ModuleInferenceFailure
        name = module.__name__
        if not name.startswith( f"{__.package_name}." ): break
        frame = frame.f_back
    return name


def _iterate_address_ancestry( name: str ) -> __.cabc.Iterator[ str ]:
    parts = name.split( '.' )
    for i in range( len( parts ) ):
        yield '.'.join( parts[ : i + 1 ] )


def _merge_ic_configuration(
    base: dict[ str, __.typx.Any ], update_objct: object,
) -> dict[ str, __.typx.Any ]:
    update: dict[ str, __.typx.Any ] = _dict_from_dataclass( update_objct )
    result: dict[ str, __.typx.Any ] = { }
    result[ 'flavors' ] = (
            dict( base.get( 'flavors', dict( ) ) )
        |   dict( update.get( 'flavors', dict( ) ) ) )
    for ename in ( 'compositor_factory', 'introducer' ):
        uvalue = update.get( ename )
        if uvalue is not None: result[ ename ] = uvalue
        elif ename in base: result[ ename ] = base[ ename ]
    return result


def _produce_ic_configuration(
    dispatcher: Dispatcher, address: str, flavor: _flavors.Flavor
) -> __.immut.Dictionary[ str, __.typx.Any ]:
    fconfigs: list[ _cfg.FlavorConfiguration ] = [ ]
    dconfig = dispatcher.generalcfg
    configd: dict[ str, __.typx.Any ] = {
        field.name: getattr( dconfig, field.name )
        for field in __.dcls.fields( dconfig )
        if not field.name.startswith( '_' ) }
    if flavor in dconfig.flavors:
        fconfigs.append( dconfig.flavors[ flavor ] )
    for address_ in _iterate_address_ancestry( address ):
        if address_ not in dispatcher.addresscfgs: continue
        mconfig = dispatcher.addresscfgs[ address_ ]
        configd = _merge_ic_configuration( configd, mconfig )
        if flavor in mconfig.flavors:
            fconfigs.append( mconfig.flavors[ flavor ] )
    if not fconfigs: raise _exceptions.FlavorInavailability( flavor )
    # Apply collected flavor configs after general and address configs.
    # (Applied in top-down order for correct overrides.)
    for fconfig in fconfigs:
        configd = _merge_ic_configuration( configd, fconfig )
    return __.immut.Dictionary( configd )


def _resolve_printer(
    factory: _printers.PrinterFactoryUnion,
    address: str,
    flavor: _flavors.Flavor,
) -> _printers.Printer:
    from .standard import Printer
    if isinstance( factory, __.io.TextIOBase ):
        return Printer( target = factory )
    return factory( address, flavor )


def _resolve_printers(
    factories: _printers.PrinterFactoriesUnion,
    address: str,
    flavor: _flavors.Flavor,
) -> _printers.Printers:
    return tuple(
        _resolve_printer( factory, address, flavor )
        for factory in factories )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Printers, printer factories, and auxiliary functions and types. '''


import colorama as _colorama

from . import __
from . import exceptions as _exceptions
from . import flavors as _flavors
from . import records as _records


_validate_arguments = (
    __.validate_arguments(
        globalvars = globals( ),
        errorclass = _exceptions.ArgumentClassInvalidity ) )


ColumnsMaxCalculator: __.typx.TypeAlias = __.typx.Annotated[
    __.typx.Union[
        __.typx.Optional[ int ],
        __.cabc.Callable[ [ ], __.typx.Optional[ int ] ],
    ],
    __.typx.Doc(
        ''' Available line length of target character screen.

            * May be an integer.
            * May be ``None`` if indeterminable or irrelevant.
            * May be a callable which takes no arguments and returns ``None``
              or an integer. This support terminal resizing, for example.
        ''' ),
]


class TextualizationControl( __.immut.DataclassObject ):
    ''' Contextual data for compositor and introducer factories. '''

    charset: __.typx.Annotated[
        __.typx.Optional[ str ],
        __.typx.Doc(
            ''' Character set encoding of target.

                May be ``None`` if indeterminable or irrelevant. ''' ),
    ] = None
    colorize: __.typx.Annotated[
        bool, __.typx.Doc( ''' Colorize textualization? ''' )
    ] = False
    columns_max_calculator: ColumnsMaxCalculator = None

    @property
    def columns_max( self ) -> __.typx.Optional[ int ]:
        ''' Available line length (maximum columns) of target.

            May be ``None`` if indeterminable or irrelevant.
        '''
        calculator = self.columns_max_calculator
        return calculator( ) if callable( calculator ) else calculator


class Printer(
    __.immut.DataclassProtocol, __.typx.Protocol,
    decorators = ( __.typx.runtime_checkable, ),
):
    ''' Abstract base class for printers. '''

    @__.abc.abstractmethod
    def __call__( self, record: str | _records.Record ) -> None:
        ''' Prints record to destination. '''
        raise NotImplementedError

    @__.abc.abstractmethod
    def provide_textualization_control(
        self
    ) -> __.typx.Optional[ TextualizationControl ]:
        ''' Provides control object for textualization, if capable. '''
        raise NotImplementedError

    # TODO: print (same as __call__)
    # TODO: print_async


Printers: __.typx.TypeAlias = __.cabc.Sequence[ Printer ]
PrinterFactory: __.typx.TypeAlias = (
    __.cabc.Callable[ [ str, _flavors.Flavor ], Printer ] )
PrinterFactoryUnion: __.typx.TypeAlias = __.io.TextIOBase | PrinterFactory
PrinterFactoriesUnion: __.typx.TypeAlias = (
    __.cabc.Sequence[ PrinterFactoryUnion ] )


@_validate_arguments
def count_columns_visual( text: str ) -> int:
    # Note: If CSI ED ("Erase on Display") or EL ("Erase in Line") sequences
    #       are used within the text, then the count will not be accurate.
    text_no_ansi = remove_ansi_c1_sequences( text )
    return __.wcwidth.wcswidth( text_no_ansi )


@_validate_arguments
def remove_ansi_c1_sequences( text: str ) -> str:
    # https://stackoverflow.com/a/14693789/14833542
    regex = __.re.compile( r'''\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])''' )
    return regex.sub( '', text )


@_validate_arguments
def produce_columns_max_calculator(
    target: __.io.TextIOBase
) -> ColumnsMaxCalculator:
    fileno_revealer = getattr( target, 'fileno', None )
    if fileno_revealer is None: return None
    try: fileno = fileno_revealer( )
    except ( IOError, OSError, __.io.UnsupportedOperation ): return None
    if not __.os.isatty( fileno ): return None

    def calculate( ) -> __.typx.Optional[ int ]:
        try: size = __.shutil.get_terminal_size( fileno )
        except Exception: return None
        return size.columns

    return calculate


@_validate_arguments
def produce_printer_factory_default(
    target: __.io.TextIOBase,
    force_color: bool = False,
) -> PrinterFactory:
    ''' Produces default printer factory associated with a stream.

        Can optionally force ANSI SGR sequences (terminal color attributes,
        etc...) on target stream.
    '''
    def produce_printer( address: str, flavor: _flavors.Flavor ) -> Printer:
        from .standard import Printer
        match __.sys.platform:
            case 'win32':
                winansi = _colorama.AnsiToWin32( target ) # pyright: ignore
                target_ = ( # pragma: no cover
                    winansi.stream if winansi.convert else target )
            case _: target_ = target
        return Printer(
            target = target_, force_color = force_color ) # pyright: ignore

    return produce_printer


# def truncate_visual( text: str, columns_max: int ) -> str:
#     lsize = 0
#     for i, c in enumerate( text ):
#         csize = __.wcwidth.wcwidth( c )
#         csize = max( 0, csize )  # control or combining character
#         if lsize + csize > columns_max:
#             # TODO? Add ellipsis.
#             return text[ : i ]
#         lsize += csize
#     return text



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Conversion of objects to lines of text. '''


from . import __
from . import core as _core


class Linearizer( __.Linearizer ):

    configuration: __.typx.Annotated[
        _core.LinearizerConfiguration,
        __.ddoc.Doc( ''' Default behaviors for textual linearizer. ''' ),
    ] = __.dcls.field( default_factory = _core.LinearizerConfiguration )

    def __call__(
        self,
        control: __.TextualizationControl,
        entity: object,
        columns_max: __.Absential[ int ] = __.absent,
    ) -> tuple[ str, ... ]:
        auxdata = _core.LinearizerState.from_configuration(
            configuration = self.configuration, control = control )
        return linearize_omni( auxdata, entity, columns_max = columns_max )


def linearize_exception_plain(
    auxdata: _core.LinearizerState,
    exception: BaseException,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    ecfg = auxdata.configuration.exceptionscfg
    tbe = __.tb.TracebackException.from_exception( exception )
    lines = [ *ecfg.interpolate( exception ) ]
    if ecfg.enable_stacktraces:
        lines.extend(
            linearize_stacktrace_plain( auxdata, tbe.stack, columns_max ) )
    # TODO: Process '__cause__' and '__context__'.
    # TODO: Process exception groups.
    return tuple( lines )


def linearize_exception_rich(
    auxdata: _core.LinearizerState,
    exception: BaseException,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    # TODO: Ensure that exception groups are handled properly.
    ecfg = auxdata.configuration.exceptionscfg
    capture = __.io.StringIO( )
    console = __.produce_rich_console( auxdata.control, capture, columns_max )
    if not ecfg.enable_stacktraces:
        console.print( exception )
        return tuple( capture.getvalue( ).split( '\n' ) )
    traceback = __.rich_traceback.Traceback.from_exception(
        type( exception ), exception, exception.__traceback__ )
    console.print( traceback )
    return tuple( capture.getvalue( ).split( '\n' ) )


def linearize_object_plain(
    auxdata: _core.LinearizerState,
    entity: object,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    # TODO? Pass configurable indentation width.
    text = (
        __.pprint.saferepr( entity ) if __.is_absent( columns_max )
        else __.pprint.pformat( entity, indent = 2, width = columns_max ) )
    return tuple( text.split( '\n' ) )


def linearize_object_rich(
    auxdata: _core.LinearizerState,
    entity: object,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    capture = __.io.StringIO( )
    console = __.produce_rich_console( auxdata.control, capture, columns_max )
    console.print( entity )
    return tuple( capture.getvalue( ).split( '\n' ) )


def linearize_omni(
    auxdata: _core.LinearizerState,
    entity: object,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    if auxdata.colorize:
        return linearize_omni_rich( auxdata, entity, columns_max )
    return linearize_omni_plain( auxdata, entity, columns_max )


def linearize_omni_plain(
    auxdata: _core.LinearizerState,
    entity: object,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    if isinstance( entity, str ):
        return linearize_text_plain( auxdata, entity, columns_max )
    if isinstance( entity, BaseException ):
        return linearize_exception_plain( auxdata, entity, columns_max )
    return linearize_object_plain( auxdata, entity, columns_max )


def linearize_omni_rich(
    auxdata: _core.LinearizerState,
    entity: object,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    if isinstance( entity, str ):
        return linearize_text_rich( auxdata, entity, columns_max )
    if isinstance( entity, BaseException ):
        return linearize_exception_rich( auxdata, entity, columns_max )
    return linearize_object_rich( auxdata, entity, columns_max )


def linearize_stacktrace_plain(
    auxdata: _core.LinearizerState,
    stacktrace: __.tb.StackSummary,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    infinite_lines = __.is_absent( columns_max )
    lines: list[ str ] = [ ]
    for frame in stacktrace:
        filename_part = f"File '{frame.filename}'"
        lineno_part = f"line {frame.lineno}" if frame.lineno else ''
        name_part = f"in {frame.name}"
        parts = ( filename_part, lineno_part, name_part )
        address = ', '.join( filter( None, parts ) )
        address_size = len( address )
        if infinite_lines or address_size <= columns_max:
            lines.append( address )
        else:
            excess_size = address_size - columns_max - 2  # sans ', '
            parts = ( lineno_part, name_part )
            address = ', '.join( filter( None, parts ) )
            address_size = len( address )
            if excess_size <= address_size:
                lines.append( filename_part )
            else:
                excess_size = columns_max - len( frame.filename ) + 4
                filename = frame.filename[ excess_size : ]
                # TODO? Drop middle rather than start.
                lines.append( f"File '... {filename}'" )
            lines.append( address )
        if frame.line:
            line = frame.line.strip( )
            # TODO? Apply Pygments to line.
            lines_ = iter(
                linearize_text_plain(
                    auxdata, line,
                    __.absent if infinite_lines else columns_max - 4 ) )
            lines.append( "    {}".format( next( lines_ ) ) )
            lines.extend( f"      {line_}" for line_ in lines_ )
    return tuple( lines )


def linearize_stacktrace_rich(
    auxdata: _core.LinearizerState,
    stacktrace: __.tb.StackSummary,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    frames = [
        __.rich_traceback.Frame(
            frame.filename, frame.lineno or -1, frame.name, frame.line or '' )
        for frame in stacktrace ]
    stack = __.rich_traceback.Stack(
        exc_type = 'Callstack', exc_value = 'Inspection', frames = frames )
    trace = __.rich_traceback.Trace( stacks = [ stack ] )
    traceback = __.rich_traceback.Traceback( trace = trace )
    capture = __.io.StringIO( )
    console = __.produce_rich_console( auxdata.control, capture, columns_max )
    console.print( traceback )
    # TODO? Remove exception lines.
    return tuple( capture.getvalue( ).split( '\n' ) )


def linearize_text_plain(
    auxdata: _core.LinearizerState,
    text: str,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    text_no_ansi = __.remove_ansi_c1_sequences( text )
    if __.is_absent( columns_max ):
        return tuple( text_no_ansi.split( '\n' ) )
    configuration = auxdata.configuration
    incise_excesses = (
        configuration.incision_boundary
        is not _core.IncisionBoundaries.Nowhere )
    incise_naturally = (
        configuration.incision_boundary
        is _core.IncisionBoundaries.Wordsplits )
    # TODO? Account for wide characters.
    return tuple( __.textwrap.wrap(
        text_no_ansi,
        break_long_words = incise_excesses,
        break_on_hyphens = incise_naturally,
        width = columns_max ) )


def linearize_text_rich(
    auxdata: _core.LinearizerState,
    text: str,
    columns_max: __.Absential[ int ] = __.absent,
) -> tuple[ str, ... ]:
    configuration = auxdata.configuration
    text_ = __.rich_text.Text.from_ansi( text )
    infinite_lines = __.is_absent( columns_max )
    incise = (
            not infinite_lines
        and configuration.incision_boundary
            is not _core.IncisionBoundaries.Nowhere )
    capture = __.io.StringIO( )
    console = __.produce_rich_console( auxdata.control, capture, columns_max )
    console.print(
        text_,
        overflow = 'ignore' if infinite_lines else 'fold',
        no_wrap = not incise )
    return tuple( capture.getvalue( ).split( '\n' ) )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Portions of configuration hierarchy. '''


from . import __
from . import flavors as _flavors
from . import textualizers as _texts


class FlavorConfiguration( __.immut.DataclassObject ):
    ''' Per-flavor configuration. '''

    compositor_factory: __.typx.Annotated[
        __.typx.Optional[ _texts.CompositorFactory ],
        __.typx.Doc(
            ''' Factory which produces compositor callable.

                Takes address and flavor as arguments. Returns compositor.

                Default ``None`` inherits from cumulative configuration.
            ''' ),
    ] = None


FlavorsRegistry: __.typx.TypeAlias = (
    __.immut.Dictionary[ _flavors.Flavor, FlavorConfiguration ] )
FlavorsRegistryLiberal: __.typx.TypeAlias = (
    __.cabc.Mapping[ _flavors.Flavor, FlavorConfiguration ] )


def produce_flavors_default( ) -> FlavorsRegistry:
    ''' Produces registry of all standard flavors without customization. '''
    flavors: FlavorsRegistryLiberal = { }
    for name, spec in _flavors.flavor_specifications_standard.items( ):
        tfactory = _texts.produce_compositor_factory_default(
            introducer = f"{spec.label}| ", trace_exceptions = spec.stack )
        flavors[ name ] = FlavorConfiguration(
            compositor_factory = tfactory )
    for alias, name in _flavors.flavor_aliases_standard.items( ):
        flavors[ alias ] = flavors[ name ]
    for level in range( 10 ):
        indent_i = '  ' * level
        indent_s = '  ' * ( level + 1 )
        tfactory = _texts.produce_compositor_factory_default(
            introducer = f"TRACE{level}| ",
            line_prefix_initial = indent_i,
            line_prefix_subsequent = indent_s )
        flavors[ level ] = FlavorConfiguration(
            compositor_factory = tfactory )
    return __.immut.Dictionary( flavors )


class AddressConfiguration( __.immut.DataclassObject ):
    ''' Per-address configuration. '''

    compositor_factory: __.typx.Annotated[
        __.typx.Optional[ _texts.CompositorFactory ],
        __.typx.Doc(
            ''' Factory which produces compositor callable.

                Takes address and flavor as arguments. Returns compositor.

                Default ``None`` inherits from cumulative configuration.
            ''' ),
    ] = None
    flavors: __.typx.Annotated[
        FlavorsRegistry,
        __.typx.Doc(
            ''' Registry of flavor identifiers to configurations. ''' ),
    ] = __.dcls.field( default_factory = FlavorsRegistry )


class DispatcherConfiguration( __.immut.DataclassObject ):
    ''' Per-dispatcher configuration. '''

    compositor_factory: __.typx.Annotated[
        _texts.CompositorFactory,
        __.typx.Doc(
            ''' Factory which produces compositor callable.

                Takes address and flavor as arguments. Returns compositor.
            ''' ),
    ] = _texts.produce_compositor_factory_default( )
    flavors: __.typx.Annotated[
        FlavorsRegistry,
        __.typx.Doc(
            ''' Registry of flavor identifiers to configurations. ''' ),
    ] = __.dcls.field( default_factory = produce_flavors_default )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Inspection of variables as function arguments. '''

# TODO: Implement if deeper inspection than f-string {name=} is needed.


# import executing as _executing


from . import __


class Inspection( __.immut.DataclassObject ):
    ''' Result of variable inspection. '''

    name: str
    value: __.typx.Any


Inspections: __.typx.TypeAlias = __.cabc.Sequence[ Inspection ]


def inspect_variables( *variables: __.typx.Any ) -> Inspections:
    ''' Returns values of variables with names and execution context. '''
    inspections: list[ Inspection ] = [ ]
    # TODO: Implement.
    return tuple( inspections )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Reporter flavors. '''


from . import __


Flavor: __.typx.TypeAlias = int | str


class StandardFlavorSpecification( __.immut.DataclassObject ):
    ''' Specification for standard flavor. '''

    color: __.typx.Annotated[
        str, __.ddoc.Doc( ''' Name of introduction color. ''' ) ]
    emoji: __.typx.Annotated[ str, __.ddoc.Doc( ''' Introduction emoji. ''' ) ]
    label: __.typx.Annotated[ str, __.ddoc.Doc( ''' Introduction label. ''' ) ]
    stack: __.typx.Annotated[
        bool, __.ddoc.Doc( ''' Include stack trace? ''' )
    ] = False


flavor_aliases_standard: __.immut.Dictionary[
    str, str
] = __.immut.Dictionary( {
    'n': 'note', 'm': 'monition',
    'e': 'error', 'a': 'abort',
    'ex': 'errorx', 'ax': 'abortx',
    'f': 'future', 's': 'success',
    'v': 'advice',
} )

flavor_specifications_standard: __.immut.Dictionary[
    str, StandardFlavorSpecification
] = __.immut.Dictionary(
    note = StandardFlavorSpecification(
        color = 'blue',
        emoji = '\N{Information Source}\ufe0f',
        label = 'NOTE' ),
    monition = StandardFlavorSpecification(
        color = 'yellow',
        emoji = '\N{Warning Sign}\ufe0f',
        label = 'MONITION' ),
    error = StandardFlavorSpecification(
        color = 'red', emoji = '❌', label = 'ERROR' ),
    errorx = StandardFlavorSpecification(
        color = 'red', emoji = '❌', label = 'ERROR', stack = True ),
    abort = StandardFlavorSpecification(
        color = 'bright_red', emoji = '💥', label = 'ABORT' ),
    abortx = StandardFlavorSpecification(
        color = 'bright_red', emoji = '💥', label = 'ABORT', stack = True ),
    future = StandardFlavorSpecification(
        color = 'magenta', emoji = '🔮', label = 'FUTURE' ),
    success = StandardFlavorSpecification(
        color = 'green', emoji = '✅', label = 'SUCCESS' ),
    advice = StandardFlavorSpecification(
        color = 'cyan', emoji = '💡', label = 'ADVICE' ),
)



# Copied from dedicated project to avoid additional dependency:
#   https://github.com/trim21/wcwidth-stubs/blob/master/src/wcwidth-stubs/__init__.pyi
def list_versions() -> tuple[str, ...]: ...
def wcwidth(wc: str, unicode_version: str = ...) -> int: ...
def wcswidth(
    pwcs: str, n: int | None = None, unicode_version: str = ...
) -> int: ...

__all__ = ("list_versions", "wcswidth", "wcwidth")



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Standard compositor and renderers. '''


from . import __
from . import core as _core

from .linearizers import linearize_omni as _linearize_omni


class Compositor( __.Compositor ):
    ''' Standard compositor. '''

    configuration: __.typx.Annotated[
        _core.CompositorConfiguration,
        __.ddoc.Doc( ''' Default behaviors and format for text. ''' ),
    ] = __.dcls.field( default_factory = _core.CompositorConfiguration )
    introducer: __.typx.Annotated[
        __.IntroducerUnion,
        __.ddoc.Doc(
            ''' String or factory which produces introduction string.

                Factory takes control object and record as arguments.
                Returns introduction string.
            ''' ),
    ] = f"{__.package_name}| "

    def __call__(
        self, control: __.TextualizationControl, record: __.Record
    ) -> str:
        configuration = self.configuration
        ecfg = configuration.linearizercfg.exceptionscfg
        auxdata = _core.CompositorState.from_configuration(
            configuration = configuration, control = control )
        content = record.content
        introducer = self.introducer
        introduction = (
            introducer if isinstance( introducer, str )
            else introducer(
                control, record, auxdata.linearizer.columns_max ) )
        if isinstance( content, __.MessageContent ):
            summary_ = content.summary
            exception = (
                None if isinstance( summary_, BaseException )
                else ecfg.discover( ) )
            summary = _render_summary( auxdata, introduction, summary_ )
            details = tuple(
                _render_detail( auxdata, detail )
                for detail in filter( None, ( exception, *content.details ) ) )
            return configuration.details_separator.join( (
                summary, *details ) )
        raise __.ContentMisclassification( type( content ) )


def _calculate_ccount_max(
    initial: str, subsequent: __.typx.Optional[ str ]
) -> int:
    i_ccount = __.count_columns_visual( initial )
    if subsequent is None: return i_ccount
    return max( i_ccount, __.count_columns_visual( subsequent ) )


def _render_detail( auxdata: _core.CompositorState, detail: object ) -> str:
    configuration = auxdata.configuration
    columns_max = auxdata.linearizer.columns_max
    detail_prefix_i = configuration.detail_prefix_initial
    detail_prefix_i_ccount = __.count_columns_visual( detail_prefix_i )
    detail_prefix_s = configuration.detail_prefix_subsequent
    if detail_prefix_s is None:
        detail_prefix_s = ' ' * detail_prefix_i_ccount
    detail_prefix_ccount = _calculate_ccount_max(
        detail_prefix_i, detail_prefix_s )
    line_prefix_ccount = _calculate_ccount_max(
        configuration.line_prefix_initial,
        configuration.line_prefix_subsequent )
    prefix_ccount = line_prefix_ccount + detail_prefix_ccount
    match auxdata.linearizer.columns_constraint:
        case _core.ColumnsConstraints.Complect:
            remainder_ccount = (
                __.absent if __.is_absent( columns_max )
                else columns_max - prefix_ccount )
        case _core.ColumnsConstraints.Exceed:
            remainder_ccount = __.absent
    lines = iter( _linearize_omni(
        auxdata.linearizer, detail, remainder_ccount ) )
    lines_final: list[ str ] = [ ]
    line_i = next( lines )
    _update_lines_collection(
        configuration, lines_final,
        f"{detail_prefix_i}{line_i}",
        tuple( f"{detail_prefix_s}{line}" for line in lines ) )
    return '\n'.join( lines_final )


def _render_summary(
    auxdata: _core.CompositorState, introduction: str, summary: object
) -> str:
    match auxdata.linearizer.columns_constraint:
        case _core.ColumnsConstraints.Complect:
            return _complect_render_summary( auxdata, introduction, summary )
        case _core.ColumnsConstraints.Exceed:
            return _exceed_render_summary( auxdata, introduction, summary )


def _complect_render_summary(
    auxdata: _core.CompositorState, introduction: str, summary: object
) -> str:
    configuration = auxdata.configuration
    columns_max = auxdata.linearizer.columns_max
    line_prefix_i = configuration.line_prefix_initial
    intro_ccount = __.count_columns_visual( introduction )
    prefix_ccount = _calculate_ccount_max(
        line_prefix_i, configuration.line_prefix_subsequent )
    remainder_ccount = (
        __.absent if __.is_absent( columns_max )
        else columns_max - prefix_ccount )
    lines_final: list[ str ] = [ ]
    lines = _linearize_omni( auxdata.linearizer, summary, remainder_ccount )
    match len( lines ):
        case 0: raise __.SummaryLinearizationFailure( )
        case 1:
            content = lines[ 0 ]
            incision_point = 0
            if not __.is_absent( columns_max ):
                incision_point = (
                        configuration.summary_incision_ratio * columns_max )
            isolate_introduction = incision_point <= intro_ccount
            if not isolate_introduction:
                candidate = f"{introduction} {content}"
                candidate_ccount = (
                        prefix_ccount + intro_ccount
                    +   __.count_columns_visual( content ) + 1 )
                if candidate_ccount <= columns_max:
                    lines_final.append( f"{line_prefix_i}{candidate}" )
                else:
                    _update_lines_collection(
                        configuration, lines_final, introduction, lines )
            else:
                _update_lines_collection(
                    configuration, lines_final, introduction, lines )
        case _:
            _update_lines_collection(
                configuration, lines_final, introduction, lines )
    return '\n'.join( lines_final )


def _exceed_render_summary(
    auxdata: _core.CompositorState, introduction: str, summary: object
) -> str:
    configuration = auxdata.configuration
    line_prefix_i = configuration.line_prefix_initial
    lines_final: list[ str ] = [ ]
    lines = _linearize_omni( auxdata.linearizer, summary )
    match len( lines ):
        case 0: raise __.SummaryLinearizationFailure( )
        case 1:
            content = lines[ 0 ]
            lines_final.append( f"{line_prefix_i}{introduction} {content}" )
        case _:
            _update_lines_collection(
                configuration, lines_final, introduction, lines )
    return '\n'.join( lines_final )


def _update_lines_collection(
    configuration: _core.CompositorConfiguration,
    collector: list[ str ],
    line_initial: str,
    lines_subsequent: __.typx.Optional[ tuple[ str, ... ] ] = None,
) -> None:
    line_prefix_i = configuration.line_prefix_initial
    line_prefix_s = configuration.line_prefix_subsequent
    if line_prefix_s is None:
        line_prefix_s = ' ' * __.count_columns_visual( line_prefix_i )
    collector.append( f"{line_prefix_i}{line_initial}" )
    if lines_subsequent:
        collector.extend(
            f"{line_prefix_s}{line}" for line in lines_subsequent )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Printers and printer factories. '''


from . import __


_validate_arguments = (
    __.validate_arguments(
        globalvars = globals( ),
        errorclass = __.ArgumentClassInvalidity ) )


class Printer( __.Printer ):
    ''' Simple printer that writes to a text stream. '''

    target: __.io.TextIOBase
    force_color: bool = False

    def __call__( self, record: str | __.Record ) -> None:
        text = record if isinstance( record, str ) else str( record )
        if not self._determine_colorization( ):
            text = __.remove_ansi_c1_sequences( text )
        print( text, file = self.target )

    def provide_textualization_control(
        self
    ) -> __.typx.Optional[ __.TextualizationControl ]:
        colorize = self._determine_colorization( )
        columns_max_calculator = (
            __.produce_columns_max_calculator( self.target ) )
        charset = getattr( self.target, 'encoding', None )
        if charset:
            charset = __.codecs.lookup( charset ).name  # normalize
        return __.TextualizationControl(
            charset = charset,
            colorize = colorize,
            columns_max_calculator = columns_max_calculator )

    def _determine_colorization( self ) -> bool:
        colorize = self.target.isatty( )
        if __.os.environ.get( 'NO_COLOR' ): colorize = False
        return colorize or self.force_color



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Internal imports for textualizers and their attendants. '''


# ruff: noqa: F401, F403, F405


from ..__ import *
from ..configuration import *
from ..exceptions import *
from ..flavors import *
from ..printers import *
from ..records import *
from ..textualizers import *

ENRICH = False
try:

    import rich.console as      rich_console
    import rich.style as        rich_style
    import rich.text as         rich_text
    import rich.traceback as    rich_traceback

    ENRICH = True  # pyright: ignore[reportConstantRedefinition]

    def produce_rich_console(
        control: TextualizationControl,
        capture: typx.IO[ str ],
        columns_max: Absential[ int ] = absent,
    ) -> rich_console.Console:
        charset = control.charset or ''
        colorize = control.colorize
        columns_max_nullable = (
            None if is_absent( columns_max ) else columns_max )
        safe = charset.startswith( 'utf-' )
        return rich_console.Console(
            file = capture,
            force_terminal = colorize,
            no_color = not colorize,
            safe_box = safe,
            width = columns_max_nullable )

except ImportError: pass



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Standard introducer with support for decorations and styles. '''


from . import __
from . import core as _core


class Introducer( __.Introducer ):
    ''' Standard introducer. '''

    configuration: __.typx.Annotated[
        _core.IntroducerConfiguration,
        __.ddoc.Doc(
            ''' Default behaviors and format for introductory text. ''' ),
    ] = __.dcls.field( default_factory = _core.IntroducerConfiguration )

    def __call__(
        self,
        control: __.TextualizationControl,
        record: __.Record,
        columns_max: __.Absential[ int ] = __.absent,
    ) -> str:
        configuration = self.configuration
        auxdata = _core.IntroducerState.from_configuration(
            configuration = configuration,
            control = control,
            columns_max = columns_max )
        if isinstance( record.flavor, int ):
            return _render_trace_label( auxdata, record )
        return _render_nominal_label( auxdata, record )


def _render_nominal_label(
    auxdata: _core.IntroducerState, record: __.Record
) -> str:
    configuration = auxdata.configuration
    styles = dict( configuration.styles )
    flavor = record.flavor
    if isinstance( flavor, int ):
        raise __.FlavorMisclassification( flavor, expectation = 'string' )
    name = __.flavor_aliases_standard.get( flavor, flavor )
    spec = __.flavor_specifications_standard[ name ]
    label = ''
    if configuration.label_as & _core.LabelPresentations.Emoji:
        if configuration.label_as & _core.LabelPresentations.Words:
            label = f"{spec.emoji} {spec.label}"
        else: label = f"{spec.emoji}"
    elif configuration.label_as & _core.LabelPresentations.Words:
        label = f"{spec.label}"
    if auxdata.colorize:
        styles[ 'flavor' ] = _core.Style( fgcolor = spec.color )
    return _render_common( auxdata, record, styles, label )


def _render_trace_label(
    auxdata: _core.IntroducerState, record: __.Record
) -> str:
    # TODO? Option to render indentation guides.
    configuration = auxdata.configuration
    styles = dict( configuration.styles )
    flavor = record.flavor
    if not isinstance( flavor, int ):
        raise __.FlavorMisclassification( flavor, expectation = 'int' )
    level = flavor
    label = ''
    if configuration.label_as & _core.LabelPresentations.Emoji:
        if configuration.label_as & _core.LabelPresentations.Words:
            label = f"🔎 TRACE{level}"
        else: label = '🔎'
    elif configuration.label_as & _core.LabelPresentations.Words:
        label = f"TRACE{level}"
    if auxdata.colorize and level < len( _trace_color_names ):
        styles[ 'flavor' ] = (
            _core.Style( fgcolor = _trace_color_names[ level ] ) )
    return _render_common( auxdata, record, styles, label )


def _render_common(
    auxdata: _core.IntroducerState,
    record: __.Record,
    styles: __.cabc.Mapping[ str, _core.Style ],
    label: str
) -> str:
    # TODO? Performance optimization: Only compute and interpolate PID, thread,
    #       and timestamp, if capabilities set permits.
    configuration = auxdata.configuration
    auxiliaries = configuration.auxiliaries
    thread = auxiliaries.thread_discoverer( )
    interpolants: dict[ str, str ] = {
        'flavor': label,
        'address': record.address,
        'timestamp': auxiliaries.time_formatter( configuration.ts_format ),
        'process_id': str( auxiliaries.pid_discoverer( ) ),
        'thread_id': str( thread.ident ),
        'thread_name': thread.name,
    }
    if auxdata.colorize:
        _stylize_interpolants( auxdata, interpolants, styles )
    return configuration.template.format( **interpolants )


def _stylize_interpolants(
    auxdata: _core.IntroducerState,
    interpolants: dict[ str, str ],
    styles: __.cabc.Mapping[ str, _core.Style ],
) -> None:
    style_default = styles.get( 'flavor' )
    interpolants_: dict[ str, str ] = { }
    for iname, ivalue in interpolants.items( ):
        style = styles.get( iname, style_default )
        if not style: continue # pragma: no branch
        capture = __.io.StringIO( )
        console = __.produce_rich_console(
            auxdata.control, capture, auxdata.columns_max )
        style_ = __.rich_style.Style( color = style.fgcolor )
        console.print( ivalue, end = '', highlight = False, style = style_  )
        interpolants_[ iname ] = capture.getvalue( )
    interpolants.update( interpolants_ )


_trace_color_names: tuple[ str, ... ] = (
    'grey85', 'grey82', 'grey78', 'grey74', 'grey70',
    'grey66', 'grey62', 'grey58', 'grey54', 'grey50' )

_trace_prefix_styles: tuple[ _core.Style, ... ] = tuple(
    _core.Style( fgcolor = name ) for name in _trace_color_names )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Records to inscribe. '''


from . import __
from . import flavors as _flavors


MessageDetail: __.typx.TypeAlias = str
MessageDetails: __.typx.TypeAlias = tuple[ MessageDetail, ... ]
MessageSummary: __.typx.TypeAlias = str | BaseException


class Content( __.immut.DataclassObject ):
    ''' Abstract base class for content. '''


class MessageContent( Content ):
    ''' Content for standard messages. '''

    summary: MessageSummary
    details: MessageDetails


class Record( __.immut.DataclassObject ):
    ''' Content with metadata. '''

    address: str
    content: Content
    flavor: _flavors.Flavor
    # TODO? 'ttl' for printing to other reporters
    ctime: __.Datetime = __.dcls.field(
        default_factory = lambda: __.Datetime.now( __.Timezone.utc ) )
    # TODO? 'excinfo' (exception info)



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Core data structures and utilities. '''


from . import __


class Auxiliaries( __.immut.DataclassObject ):
    ''' Auxiliary functions used by textualizers and interpolation.

        Typically used by unit tests to inject mock dependencies,
        but can also be used to deeply customize output.
    '''

    pid_discoverer: __.typx.Annotated[
        __.typx.Callable[ [ ], int ],
        __.ddoc.Doc( ''' Returns ID of current process. ''' ),
    ] = __.os.getpid
    thread_discoverer: __.typx.Annotated[
        __.typx.Callable[ [ ], __.threads.Thread ],
        __.ddoc.Doc( ''' Returns current thread. ''' ),
    ] = __.threads.current_thread
    time_formatter: __.typx.Annotated[
        __.typx.Callable[ [ str ], str ],
        __.ddoc.Doc( ''' Returns current time in specified format. ''' ),
    ] = lambda fmt: __.Datetime.now( __.Timezone.utc ).strftime( fmt )


class ColumnsConstraints( __.enum.Enum ):
    ''' How to constrain text which exceeds maximum columns. '''

    Complect    = __.enum.auto( )  # fold/wrap
    Exceed      = __.enum.auto( )  # overflow
    # Truncate    = __.enum.auto( )  # chop/cut


class IncisionBoundaries( __.enum.Enum ):
    ''' Where to constrain text which exceeds maximum columns. '''

    Nowhere     = __.enum.auto( )
    Whitespace  = __.enum.auto( )  # horizontal spaces and tabs
    Wordsplits  = __.enum.auto( )  # hyphens + whitespace
    Anywhere    = __.enum.auto( )


class Style( __.immut.DataclassObject ):
    ''' Style for text. Corresponds to terminal attributes. '''

    bgcolor: __.typx.Optional[ str ] = None
    fgcolor: __.typx.Optional[ str ] = None
    # TODO: Int flag enum for bold, blink, etc...


InterpolantsStylesRegistry: __.typx.TypeAlias = (
    __.accret.Dictionary[ str, Style ] )


class LabelPresentations( __.enum.IntFlag ):
    ''' How introduction labels should be presented. '''

    Nothing =   0
    Words =     __.enum.auto( )
    Emoji =     __.enum.auto( )


class ExceptionsConfiguration( __.immut.DataclassObject ):
    ''' Configuration pertaining to exceptions. '''

    discoverer: __.typx.Annotated[
        __.typx.Callable[ [ ], __.ExceptionInfo ],
        __.ddoc.Doc( ''' Returns information on current exception. ''' ),
    ] = __.sys.exc_info
    enable_discovery: __.typx.Annotated[
        bool, __.ddoc.Doc( ''' Discover active exception? ''' )
    ] = False
    enable_stacktraces: __.typx.Annotated[
        bool, __.ddoc.Doc( ''' Render tracebacks? ''' )
    ] = False
    recursive_stacktraces: __.typx.Annotated[
        bool, __.ddoc.Doc(
            ''' Render traceback for each exception group member? ''' ),
    ] = False
    template: __.typx.Annotated[
        str, __.ddoc.Doc( ''' Template for exception message. ''' )
    ] = '[{name}] {message}'

    def discover( self ) -> __.typx.Optional[ BaseException ]:
        ''' Discovers active exception. '''
        return self.discoverer( )[ 1 ] if self.enable_discovery else None

    def interpolate( self, exception: BaseException ) -> tuple[ str, ... ]:
        ''' Interpolates exception attributes into message template. '''
        eclass = type( exception )
        name = eclass.__name__
        qname = eclass.__qualname__
        mname = eclass.__module__
        interpolants = dict(
            name = name, qname = qname, mname = mname,
            message = str( exception ) )
        interpolants[ 'fqname' ] = f"{mname}.{qname}"
        return tuple( self.template.format( **interpolants ).split( '\n' ) )


class IntroducerConfiguration( __.immut.DataclassObject ):
    ''' Behaviors and format for text from standard introducer. '''

    auxiliaries: __.typx.Annotated[
        Auxiliaries, __.typx.Doc( ''' Auxiliaries for interpolation. ''' )
    ] = __.dcls.field( default_factory = Auxiliaries )
    colorize: __.typx.Annotated[
        bool, __.typx.Doc( ''' Attempt to colorize? ''' )
    ] = True
    label_as: __.typx.Annotated[
        LabelPresentations,
        __.ddoc.Doc(
            ''' How to present prefix label.

                ``Words``: As words like ``TRACE0`` or ``ERROR``.
                ``Emoji``: As emoji like ``🔎`` or ``❌``.

                For both emoji and words: ``Emoji | Words``.
            ''' )
    ] = LabelPresentations.Words
    styles: __.typx.Annotated[
        InterpolantsStylesRegistry,
        __.ddoc.Doc(
            ''' Mapping of interpolant names to style objects. ''' ),
    ] = __.dcls.field( default_factory = InterpolantsStylesRegistry )
    template: __.typx.Annotated[
        str,
        __.ddoc.Doc(
            ''' String format for prefix.

                The following interpolants are supported:
                ``flavor``: Decorated flavor.
                ``address``: Address of invoker.
                ``timestamp``: Current timestamp, formatted as string.
                ``process_id``: ID of current process according to OS kernel.
                ``thread_id``: ID of current thread.
                ``thread_name``: Name of current thread.
            ''' ),
    ] = "{flavor}| " # "{timestamp} [{module_qname}] {flavor}| "
    ts_format: __.typx.Annotated[
        str,
        __.ddoc.Doc(
            ''' String format for prefix timestamp.

                Used by :py:func:`time.strftime` or equivalent.
            ''' ),
    ] = '%Y-%m-%d %H:%M:%S.%f'


class IntroducerState( __.immut.DataclassObject ):
    ''' Data transfer object for introducer state. '''

    configuration: IntroducerConfiguration
    control: __.TextualizationControl
    colorize: __.typx.Annotated[ bool, __.ddoc.Doc( ''' Colorize? ''' ) ]
    columns_max: __.typx.Annotated[
        __.Absential[ int ],
        __.ddoc.Doc(
            ''' Available line length (maximum columns) of target. ''' ),
    ] = __.absent

    @classmethod
    def from_configuration(
        cls,
        configuration: IntroducerConfiguration,
        control: __.TextualizationControl,
        columns_max: __.Absential[ int ] = __.absent,
    ) -> __.typx.Self:
        colorize = __.ENRICH and control.colorize and configuration.colorize
        return cls(
            configuration = configuration,
            control = control,
            colorize = colorize,
            columns_max = columns_max )


class LinearizerConfiguration( __.immut.DataclassObject ):
    ''' Behaviors for standard textual linearizer. '''

    colorize: __.typx.Annotated[
        bool, __.typx.Doc( ''' Attempt to colorize? ''' )
    ] = True
    columns_constraint: __.typx.Annotated[
        ColumnsConstraints,
        __.ddoc.Doc(
            ''' How to constrain text which exceeds maximum columns. ''' ),
    ] = ColumnsConstraints.Complect
    columns_max: __.typx.Annotated[
        __.typx.Optional[ int ],
        __.ddoc.Doc(
            ''' How many columns per line to assume if printer does not tell.

                If ``None``, then infinite number of columns is assumed.
            ''' ),
    ] = None
    exceptionscfg: __.typx.Annotated[
        ExceptionsConfiguration,
        __.ddoc.Doc( ''' Configuration pertaining to exceptions. ''' ),
    ] = __.dcls.field( default_factory = ExceptionsConfiguration )
    incision_boundary: __.typx.Annotated[
        IncisionBoundaries,
        __.ddoc.Doc(
            ''' Where to constrain text which exceeds maximum columns. ''' ),
    ] = IncisionBoundaries.Wordsplits


class LinearizerState( __.immut.DataclassObject ):
    ''' Data transfer object for linearizer state. '''

    configuration: LinearizerConfiguration
    control: __.TextualizationControl
    colorize: __.typx.Annotated[ bool, __.ddoc.Doc( ''' Colorize? ''' ) ]
    columns_constraint: __.typx.Annotated[
        ColumnsConstraints,
        __.ddoc.Doc( ''' Effective columns constraint for lines. ''' ),
    ] = ColumnsConstraints.Exceed
    columns_max: __.typx.Annotated[
        __.Absential[ int ],
        __.ddoc.Doc(
            ''' Available line length (maximum columns) of target. ''' ),
    ] = __.absent

    @classmethod
    def from_configuration(
        cls,
        configuration: LinearizerConfiguration,
        control: __.TextualizationControl,
    ) -> __.typx.Self:
        colorize = __.ENRICH and control.colorize and configuration.colorize
        columns_constraint = configuration.columns_constraint
        columns_max = control.columns_max or configuration.columns_max
        if columns_max is None:
            columns_constraint = ColumnsConstraints.Exceed
            columns_max = __.absent
        return cls(
            configuration = configuration,
            control = control,
            colorize = colorize,
            columns_constraint = columns_constraint,
            columns_max = columns_max )


class CompositorConfiguration( __.immut.DataclassObject ):
    ''' Behaviors and format for text from standard compositor. '''

    detail_prefix_initial: __.typx.Annotated[
        str, __.ddoc.Doc( ''' Initial prefix for message detail. ''' )
    ] = ''
    detail_prefix_subsequent: __.typx.Annotated[
        __.typx.Optional[ str ],
        __.ddoc.Doc(
            ''' Subsequent prefix for message detail.

                If ``None``, then automatic padding is calculated based on the
                visual width of the initial prefix for message detail.
            ''' ),
    ] = None
    # TODO? 'details_maximum'
    details_separator: __.typx.Annotated[
        str, __.ddoc.Doc( ''' Separator between details. ''' )
    ] = '\n\n'
    line_prefix_initial: __.typx.Annotated[
        str, __.ddoc.Doc( ''' Prefix before first line. ''' )
    ] = ''
    line_prefix_subsequent: __.typx.Annotated[
        __.typx.Optional[ str ],
        __.ddoc.Doc(
            ''' Prefix before each line after the first.

                If ``None``, then automatic padding is calculated based on the
                visual width of the initial line prefix.
            ''' ),
    ] = None
    linearizercfg: __.typx.Annotated[
        LinearizerConfiguration,
        __.ddoc.Doc(
            ''' Text linearization and pretty-formatting behaviors. ''' ),
    ] = __.dcls.field( default_factory = LinearizerConfiguration )
    summary_incision_ratio: __.typx.Annotated[
        float,
        __.ddoc.Doc(
            ''' Ratio of introduction width to full width at which to split.

                If ratio is met or exceeded, then introduction and summary are
                split onto consecutive lines.
            ''' ),
    ] = 0.3


class CompositorState( __.immut.DataclassObject ):
    ''' Data transfer object for textualizer state. '''

    configuration: CompositorConfiguration
    linearizer: LinearizerState

    @classmethod
    def from_configuration(
        cls,
        configuration: CompositorConfiguration,
        control: __.TextualizationControl,
    ) -> __.typx.Self:
        linearizer = LinearizerState.from_configuration(
            configuration = configuration.linearizercfg, control = control )
        return cls( configuration = configuration, linearizer = linearizer )


COMPOSITOR_CONFIGURATION_DEFAULT = CompositorConfiguration( )
INTRODUCER_CONFIGURATION_DEFAULT = IntroducerConfiguration( )
